#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <assert.h>

#ifdef QDOS
#include <qdos.h>
#include <qptr.h>
char _prog_name[] = "vplay";
int FILEBUFSIZE= 16*1024;
int USERBUFSIZE= 1024;
#else
#define strnicmp strncasecmp 
int FILEBUFSIZE= 32*1024;
int USERBUFSIZE= 16*1024;
#endif

#define MALLOCSIZE (USERBUFSIZE+FILEBUFSIZE)

unsigned char *Fbuf, *Ubuf;

enum cchar
{
    NUL, SOH, STX, ETX, EOT, ENQ, ACK, BEL, BS, HT, LF,
    VT, FF, CR, SO, SI, DLE, XON, DC2, XOFF, DC4, NAK,
    SYN, ETB, CAN, EM, SUB, ESC, FS, GS, RS, US
};

static char *ccname[32] =
{
    "NUL", "SOH", "STX", "ETX", "EOT", "ENQ", "ACK", "BEL", "BS", "HT", "LF",
    "VT", "FF", "CR", "SO", "SI", "DLE", "XON", "DC2", "XOFF", "DC4", "NAK",
    "SYN", "ETB", "CAN", "EM", "SUB", "ESC", "FS", "GS", "RS", "US"};

static char *promptstr[] = {"OK", "CONNECT", "VCON", "NO CARRIER", NULL};

static char Zhead[16] = {'Z', 'y', 'X', 'E', 'L', 2};
#ifdef __unix__
extern int tty_open (char *, long, short);
#else
extern int tty_open (char *, short);
#endif

extern int SRead(void *, int, int);
extern int SWrite(void *, int);
static short xflag = 0;

static short Head2VSM (short head)
{
    short rval = 0;

    switch (head)
    {
	case 0:
	case 1:
	case 2:
	case 3:
	    rval = head + 1;
	    break;
	case 5:
	case 6:
	case 7:
	    rval = head - 3;
	    break;
	case 8:
	case 10:
	    rval = 30;
	    break;
    }
    return rval;
}

static int faxgets (char *s, int n, int t, short arc)
{
    int c = 0;
    char *p;
    char termc = '\n';
    short tlm, tc;
    short first = 1;
    int q = 0;

    p = s;
    *p = 0;
    tlm = t * 10;
    tc = 0;
    n--;

    while (n >= 0)
    {
	c = SRead (p, 1, 1);
	if (c == 0)
	{
	    if (tc++ == tlm)
	    {
		q = -2;
		break;
	    }
	}
	else
	{
	    if (first)
	    {
		if (tolower (*p) == 'a')
		{
		    termc = '\r';
		}
		first = 0;
	    }
	    n--;
	    if (*p++ == termc)
	    {
		break;
	    }
	}
    }
    *p = 0;
    return q;
}

char *strtabmatch (char **p, char *s)
{
    while (*p && strncmp (*p, s, strlen (*p)))
	p++;
    return *p;
}

char *cmd (char *s, int t)
{
    static char buf[128];
    int p;
    *buf = 0;
    
    if (s)
    {
	if (strnicmp (s, "AT", 2) == 0)
	{
	    s += 2;
	}
	sprintf (buf, "AT%s", s);

	if (SWrite (buf, -1) == -1)
	{
	    return (char *) EOF;
	}
	SWrite("\r", 1);
    }
    fprintf(stderr, "[%s] = ", buf);
    
    while ((p = faxgets (buf, 128, t, 0)) == 0)
    {
	if (strtabmatch (promptstr, buf))
	{
	    break;
	}
    }
    fprintf(stderr, " %s", buf);
    fflush(stderr);
    
    return p == 0 ? buf : NULL;
}

short SendData (FILE *f, unsigned char *Fbuf, unsigned char *obuf, 
		int bsize, short flag)
{
    int done = 0, err = 0;
    unsigned char q;
    unsigned char *ip, *ep, *op = obuf;
    short dle = 0;
    int nc = bsize;
    short uerr = 0;
    int tc = 0;
    int oldp;
    
    ip = Fbuf;
    ep = obuf + USERBUFSIZE / 2;

#ifdef QDOS
    oldp = mt_prior(-1, 65);
#endif
    while (!done)
    {
	if (nc--)
	{
	    *op++ = *ip;
	    if (*ip == DLE)
		*op++ = DLE;
	    ip++;
	}
	else
	{
	    ip = Fbuf;
	    if ((nc = fread (Fbuf, 1, bsize, f)) <= 0)
	    {
		done = 1;
	    }
	}

	if(xflag && ((tc % 512) == 0))
	{
	    int nip;
	    while (1)
	    {
		nip  = SRead ((char *) &q, 1, 0);
		if(nip == 0)
		{
		    break;
		}

		if(nip == 1)
		{

		    if (dle)
		    {
			switch (q)
			{
			    case '0':
			    case '1':
			    case '2':
			    case '3':
			    case '4':
			    case '5':
			    case '6':
			    case '7':
			    case '8':
			    case '9':
			    case '#':
			    case '*':
			    case 'b':
			    case 'c':
			    case 'd':
			    case 'e':
			    case 'q':
			    case 's':
				err = q;
				done = 1;
				op = obuf;
			    default:
				dle = 0;
				break;
			}
		    }
		    else if (q == DLE)
		    {
			dle = 1;
		    }
		}
	    }
	}
	
	if (op >= ep || done)
	{
	    if (done && flag)
	    {
		*op++ = DLE;
		*op++ = ETX;
	    }
	    {
		int n = op - obuf;
		SWrite ((char *) obuf, n);
	    }
	    op = obuf;
	}
	tc++;
    }
#ifdef QDOS
    mt_prior(-1, oldp);
#endif
    return (uerr) ? uerr : err;
}

char *FileName = NULL;

int DumpFile (void)
{
    FILE *fp;
    int n;
    char txt[32];
    short err = 0;
    
    if ((fp = fopen (FileName, "r")))
    {
	struct stat s;
	short lmthd;
	
	fstat(fileno(fp), &s);

	fread (Fbuf, 1, 16, fp);
	sprintf (txt, "+VSM=%d", (lmthd = Head2VSM (Fbuf[10])));
	assert (*cmd (txt, 5) == 'O');

	if (!err)
	{
	    char *r;
	    n = fread (Fbuf, 1, FILEBUFSIZE, fp);
	    assert (*cmd ("+VTX", 5) == 'C');
	    {
		err = SendData (fp, Fbuf, Ubuf, n, 1);
		cmd (NULL, 10);
	    }
	}
	fclose (fp);
    }
    return err;
}

int main(int ac, char **av)
{
    int c;
    char *dev;
    char *r;
    int vol=0;
    
    while((c = getopt(ac, av, "v:f:u:d:xz")) != EOF)
    {
	switch(c)
	{
	  case 'f':
	    FILEBUFSIZE = atoi(optarg);
	    break;
	  case 'u':
	    USERBUFSIZE = 2*atoi(optarg);
	    break;
	  case 'd':
	    dev = optarg;
	    break;
	  case 'v':
	    vol = atoi(optarg);
	    break;
	  case 'z':
	    xflag = 1;
	    break;
	  case 'x':
	    {
		extern char *_endmsg;
		_endmsg = NULL;
	    }
	    break;
	}
    }
    if(dev && (FileName = av[optind]))
    {
	char p[32];
#ifdef __linux__
	tty_open(dev, 38400, 1);
#else
	tty_open(dev, 0);
#endif
	assert (*cmd("Z", 5) == 'O');
	assert (*cmd("+FCLASS=8", 5) == 'O');
	assert (*cmd("+VIP", 5) == 'O');
	assert (*cmd("+FLO=2 +VLS=16", 5) == 'V');
	if(vol)
	{
	    sprintf(p, "+VGT=%d", vol);
	    assert(*cmd(p, 5) == 'O');
	}
	
	Ubuf = (unsigned char *) calloc (MALLOCSIZE, 1);
	if (Ubuf)
	{
	    Fbuf = Ubuf + USERBUFSIZE;
	}
	DumpFile();
    }
    return 0;
}
